/* Numerical minimization */ 
/* Reinhard Oldenburg 2005, oldenburg@ph-heidelberg.de */
/* optim - a interface to fast scilab optimization */

optim:= proc(f,vars:DOM_LIST,x0:DOM_LIST)
local fx,n,ff,i,x0s,gs,gg,res;
begin
n:=nops(vars);
if not(nops(x0)=n) then error("Dimension mismatch"); end_if;
ff:=f; gg:=map(vars,v -> diff(f,v));
for i from 1 to n do ff:=subs(ff,vars[i]=scilabx(i));gg:=subs(gg,vars[i]=scilabx(i)); end_for;
fs:=expr2text(ff);
gs:="["; 
for i from 1 to n do gs:=gs.expr2text(gg[i]); if i<n then gs:=gs.";"; end_if; end_for;
gs:=gs."]";
x0s:= stringlib::subs(expr2text(x0),","=";");
fx:= scilab::func([],["xopt"],
["deff('[f,g,ind]=costf(scilabx,ind)','f=".fs.",g=".gs."')",
"[f,xopt]=optim(costf,".x0s.")" ]):
res:=fx();
if domtype(res)=DOM_LIST then res else 
map(coerce(res,DOM_LIST),e->e[1]) end_if
end_proc: 


/*---------------- */

isMinBracket:=proc(f,a,b,c)
  local fa,fb,fc;
begin
  if args(0)<4 then error("not enough arguments"); end_if;
  if domtype(f)<>DOM_PROC and domtype(f)<>DOM_FUNC_ENV then return(FAIL); end_if;
  fa:=float(f(a)); fb:=float(f(b)); fc:=float(f(c));
  return(bool(fa>fb and fc>fb));
end_proc:

nrsign:= proc(a,b) begin if float(b)>=0 then abs(a) else -abs(a) end_if; end_proc:

mnbrak:=proc(f,ax,bx)
  local fa,fb,fc,cx,ulim,u,r,q,futiny,gold,glimit,counter;
  begin
  if args(0)<3 then error("not enough arguments"); end_if;
  if domtype(f)<>DOM_PROC and domtype(f)<>DOM_FUNC_ENV then error("first operand must be of type proc"); end_if;
  gold:=1.618034; 
  glimit:=100;
  tiny:=1e-20;
  if float(ax)=float(bx) then return(mnbrak(f,ax-1,ax+1)); end_if;
  ax:=float(ax); bx:=float(bx);
  fa:=float(f(ax)); fb:=float(f(bx));
  if fa=fb then  /* Suche nach neuem b */
    counter:=0;
    while counter<100 and fa=fb do
      bx:=0.9*(bx-ax)+ax; fb:=float(f(bx)); counter:=counter+1; 
    end_while;
    if fa=fb then
      counter:=0;
      while counter<40 and fa=fb do
        bx:=2*(bx-ax)+ax; fb:=float(f(bx)); counter:=counter+1; 
      end_while;
    end_if;
    if fa=fb then print("fa=fb"); return(FAIL); end_if;
  end_if; /* Jetzt is f(ax)!=f(bx) */
  if(fb>fa) then mnbrak(f,bx,ax); else
    cx:=bx+gold*(bx-ax);
    fc:=float(f(cx));
    while fb>fc do
      r:=(bx-ax)*(fb-fc);
      q:=(bx-cx)*(fb-fa);
      u:=bx-((bx-cx)*q-(bx-ax)*r) / (2*nrsign(max(abs(q-r),tiny),q-r));
      ulim:=bx+glimit*(cx-bx);
      if (bx-u)*(u-cx)>0 then
        fu:=float(f(u));
	if fu<fc then
          ax:=bx; bx:=u; fa:=fb; fb:= fu; return([ax,bx,cx]);
	else
	  if fu>fb then
	    cx:=u; fc:=fu; return([ax,bx,cx]);
	    end_if;
	end_if;
        u:=cx+gold*(cx-bx); fu:=float(f(u));
      else 
        if (cx-u)*(u-ulim)>0 then
	  fu:=float(f(u));
  	  if fu<fc then
	    bx:=cx; cx:=u; u:=cx+gold*(cx-bx);
	    fb:=fc; fc:=fu; fu:=float(f(u)); 
	  end_if;
	else 
	  if (u-ulim)*(ulim-cx)>=0 then
	    u:=ulim; fu:=float(f(u));
	  else
	    u:=cx+gold*(cx-bx); fu:=float(f(u));
	  end_if;
        end_if;       
      end_if;
      ax:=bx; bx:=cx; cx:=u; 
      fa:=fb; fb:=fc; fc:=fu;
    end_while;
    return([ax,bx,cx]);
  end_if;
end_proc:


brentMINI:= proc(f,var,x0)
  local n,i,P,ff;
begin
  ff:= x -> subs(f,var=x); 
  brentMini(ff,x0,0.0000001)
end_proc:



brentMini:=proc(f,x0,tol)
  local br;
begin
  if args(0)<3 then error("not enough arguments"); end_if;
  if domtype(f)<>DOM_PROC and domtype(f)<>DOM_FUNC_ENV then return(FAIL); end_if;
  if x0=0 then br:=mnbrak(f,-0.1,0.1); else br:=mnbrak(f,0.99*x0,1.01*x0); end_if;
  brent(f,br[1],br[2],br[3],tol);
end_proc:

brent:= proc(f,ax,bx,cx,tol)
local cgold,zeps,itmax,iter,a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm,e;
begin
  if args(0)<5 then error("not enough arguments"); end_if;
  if domtype(f)<>DOM_PROC and domtype(f)<>DOM_FUNC_ENV then return(FAIL); end_if;
  itmax:=100;
  cgold:=0.3819660;
  zeps:=1e-10;
  e:=0.0;
  if ax<cx then a:=ax else a:=cx; end_if;
  if ax>cx then b:=ax else b:=cx; end_if;
  v:=bx; w:=v; x:=w;
  fx:=float(f(x)); fv:=fx; fw:=fv;
  for iter from 1 to itmax do
    xm:=0.5*(a+b);
    tol1:=tol*abs(x)+zeps;
    tol2:=2.0*tol1;
    if abs(x-xm)<=(tol2-0.5*(b-a)) then return(x); end_if; // done?
    if abs(e)>tol1 then
      r:=(x-w)*(fx-fv); q:=(x-v)*(fx-fw); p:=(x-v)*q-(x-w)*r; q:=2.0*(q-r); 
      if q>0 then p:=-p; end_if;
      q:=abs(q);
      etemp:=e; e:=d; 
      if abs(p)>=abs(0.5*q*etemp) or p<=q*(a-x) or p>=q*(b-x) then
        if x>=xm then e:=a-x; else e:=b-x; end_if;
        d:=cgold*e;
      else
        d:=p/q; u:=x+d; if u-a<tol2 or b-u>tol2 then d:=nrsign(tol1,d); end_if;
      end_if;
    else
      if x>=xm then e:=a-x; else e:=b-x; end_if;
      d:=cgold*e;
    end_if;
    if abs(d)>tol1 then u:=x+d; else u:=x+nrsign(tol1,d); end_if;
    fu:=float(f(u));
    if fu<=fx then 
      if u>=x then a:=x; else b:=x; end_if;
      v:=w; w:=x; x:=u;
      fv:=fw; fw:=fx; fx:=fu;
    else
      if u<x then a:=u; else b:=u; end_if;
      if fu<=fw or w=x then
        v:=w; w:=u; fv:=fw; fw:=fu;
      else
        if fu<=fv or v=x or v=w then v=u; fv:=fu; end_if;
      end_if;
    end_if;
  end_for;
  print("Warning: Max number of iterations exhaustet!");
  return(x);
end_proc:


/* P,N sind Vektoren (typ matrix), f ist Vektor-Funktion */
/* Rckgabe: [lambda,Pneu,Nneu] */
linmin:=proc(f,P,N,tol)
local ec,br,lambda, ff;
begin
  if args(0)<4 then error("not enough arguments"); end_if;
  if domtype(f)<>DOM_PROC and domtype(f)<>DOM_FUNC_ENV then return(FAIL); end_if;
  ff:= s -> float(f(P+s*N));
  ec:=traperror((br:=mnbrak(ff,0,1)));
  if ec>0 then print(["mnbrak-error trapped:",ec]); return([0,P,P-P]); end_if;
  if br=FAIL then return(FAIL); end_if;
  lambda:=brent(ff,br[1],br[2],br[3],tol);
  if lambda=FAIL then return(FAIL); end_if;
  return([lambda,P+lambda*N,lambda*N]);
end_proc:


/* powell n-dim, unconstrained
  f : Funktion die Vektor (matrix([..]) als Argument nimmt
  P : Startpunkt (Vektor)
  ftol : genauigkeit 
  */

POWELL:= proc(f,vars:DOM_LIST,x0:DOM_LIST)
  local n,i,P,ff;
begin
  n:=nops(vars);
  if not(n=nops(x0)) then error("Dim mismatch"); end_if;
  ff:= xvec -> subs(f,(vars[i]=xvec[i])$i=1..n); 
  P:=matrix(x0);
  map(coerce(powell(ff,P,0.00001),DOM_LIST),e->e[1])
end_proc:


powell:= proc(f,P,ftol)
local res,iter,itmax,fret,i,ibig,j,xi,pt,del,fp,fptt,t,ptt,xit,n,tiny,k;
begin
  if args(0)<3 then error("not enough arguments"); end_if;
  if domtype(P)=DOM_LIST then P:=matrix(P); end_if;
  if domtype(ftol)<> DOM_FLOAT then ftol:=0.00001; print("powell: ftol ungeeignet"); end_if;
  if domtype(f)<>DOM_PROC and domtype(f)<>DOM_FUNC_ENV then return(FAIL); end_if;
  tiny:=1e-25;
  itmax:=200;
  n:=nops(P);
  fret:=float(f(P));
  iter:=1;
  pt:=P;
  xi:=matrix(n,n);
  xit:=matrix(n,1);
  for i from 1 to n do for j from 1 to n do 
    if i=j then xi[i,j]:=1 else xi[i,j]:=0; end_if; 
  end_for; end_for;
  while iter>0 do
    iter:=iter+1;
    fp:=fret; ibig:=0; del:=0.0; 
    for i from 1 to n do
      for j from 1 to n do xit[j]:=xi[j,i]; end_for;
      fptt:=fret;
      res:=linmin(f,P,xit,ftol);
      if res=FAIL then next end_if;
      P:=res[2]; xit:=res[3]; fret:=float(f(P));
      if fptt-fret>del then
        del:=fptt-fret; ibig:=i;
      end_if;
    end_for;
    if 2*(fp-fret)<=ftol*(abs(fp)+abs(fret))+tiny then return(P); end_if;
    if iter>=itmax then return(FAIL); end_if;
    for j from 1 to n do
      ptt[j]:=2.0*P[j]-pt[j];
      xit[j]:=P[j]-pt[j];
      pt[j]:=P[j];
    end_for;
    fptt:=float(f(ptt));
    if fptt<fp then
      t:=2*(fp-2*fret+fptt)*(fp-fret-del)^2-del*(fp-fptt)^2;
      if t<0 then
        res:=linmin(f,P,xit,ftol); 
	if res=FAIL then next end_if;
        P:=res[2]; xit:=res[3]; fret:=float(f(P));
	for j from 1 to n do
	  xi[j,ibig]:=xi[j,n];
 	  xi[j,n]:=xit[j];
	end_for;
      end_if;
    end_if;
  end_while;
end_proc:


/* newton n-dim, unconstrained
  f : Funktion die Vektor (matrix([..]) als Argument nimmt
  P : Startpunkt (Vektor)
  ftol : genauigkeit 
  */

NEWTON:= proc(f,vars:DOM_LIST,x0:DOM_LIST)
  local n,i,P,ff;
begin
  n:=nops(vars);
  if not(n=nops(x0)) then error("Dim mismatch"); end_if;
  ff:= xvec -> subs(f,(vars[i]=xvec[i])$i=1..n); 
  P:=matrix(x0);
  map(coerce(newton(ff,P,0.00001),DOM_LIST),e->e[1])
end_proc:


newton:= proc(f,P,ftol)  
local n,i,gr,he,hei,fp,fpnew,res,iter,itmax,vars,vmat;
save v;
begin
  delete v;
  if args(0)<3 then error("not enough arguments"); end_if;
  if domtype(P)=DOM_LIST then P:=matrix(P); end_if;
  if domtype(ftol)<> DOM_FLOAT then ftol:=0.00001; print("newton: ftol ungeeignet"); end_if;
  if domtype(f)<>DOM_PROC and domtype(f)<>DOM_FUNC_ENV then return(FAIL); end_if;
  n:=nops(P);
  vars:=map([$1..n],i->v[i]);
  vmat:=matrix(vars);
  gr:=linalg::grad(f(vmat),vars);
  he:=linalg::hessian(f(vmat),vars);
  itmax:=200;
  iter:=0;
  fp:=float(f(P));
  while iter<itmax do
    iter:=iter+1;
    /* dir:= - Hessian^-1 * grad f */
    hei:=float(varsubs(he,vars,P))^(-1);
    dir:= hei*float(varsubs(gr,vars,P)); 
    res:=linmin(f,P,dir,ftol); 
    if res=FAIL then
      P:=P+10*ftol*dir; fpnew:=float(f(P));
      if fpnew<fp then fp:=fpnew; else print("newton: i-step failed!"); return(P); end_if;
    else
      if norm(P-res[2],2)<ftol then return(res[2]); end_if;
      P:=res[2]; fpnew:=float(f(P));
      if fpnew<fp then fp:=fpnew; else print("newton:step failed!"); return(P); end_if;
    end_if;
  end_while;
  if iter=itmax then print("Warning: max numer of iterations reached!"); end_if;
  return(P);
end_proc:


/* Utilities */

varsubs:=proc(term,varlist,values)
local i,r;
begin
  r:=term;
  for i from 1 to nops(varlist) do
    r:=subs(r,varlist[i]=values[i]);
  end_for;
  return(r);
end_proc:


eq2term:= e -> lhs(e)-rhs(e):
eq2term2:= e -> (lhs(e)-rhs(e))^2:

sumlist:= a -> fp::apply(_plus,op(a)): 
equatelists:= (a,b) -> {op(zip(a,b,_equal))}:


:
